@@ -16,11 +16,12 @@ import copy, os, re
 from collections import OrderedDict
 import itertools, pathlib
 import pickle
+from functools import lru_cache
 
 from . import environment
 from . import dependencies
 from . import mlog
-from .mesonlib import File, MesonException, listify, extract_as_list
+from .mesonlib import File, MesonException, listify, extract_as_list, OrderedSet
 from .mesonlib import typeslistify, stringlistify, classify_unity_sources
 from .mesonlib import get_filenames_templates_dict, substitute_values
 from .mesonlib import for_windows, for_darwin, for_cygwin, for_android, has_path_sep
@@ -71,6 +72,7 @@ buildtarget_kwargs = set([
     'objects',
     'override_options',
     'sources',
+    'gnu_symbol_visibility',
 ])
 
 known_build_target_kwargs = (
@@ -82,7 +84,7 @@ known_build_target_kwargs = (
     cs_kwargs)
 
 known_exe_kwargs = known_build_target_kwargs | {'implib', 'export_dynamic'}
-known_shlib_kwargs = known_build_target_kwargs | {'version', 'soversion', 'vs_module_defs'}
+known_shlib_kwargs = known_build_target_kwargs | {'version', 'soversion', 'vs_module_defs', 'darwin_versions'}
 known_shmod_kwargs = known_build_target_kwargs
 known_stlib_kwargs = known_build_target_kwargs | {'pic'}
 known_jar_kwargs = known_exe_kwargs | {'main_class'}
@@ -107,6 +109,10 @@ class Build:
         self.projects_args = {}
         self.global_link_args = {}
         self.projects_link_args = {}
+        self.cross_global_args = {}
+        self.cross_projects_args = {}
+        self.cross_global_link_args = {}
+        self.cross_projects_link_args = {}
         self.tests = []
         self.benchmarks = []
         self.headers = []
@@ -118,6 +124,7 @@ class Build:
         self.subproject_dir = ''
         self.install_scripts = []
         self.postconf_scripts = []
+        self.dist_scripts = []
         self.install_dirs = []
         self.dep_manifest_name = None
         self.dep_manifest = {}
@@ -167,20 +174,25 @@ class Build:
     def get_install_subdirs(self):
         return self.install_dirs
 
-    def get_global_args(self, compiler):
-        return self.global_args.get(compiler.get_language(), [])
+    def get_global_args(self, compiler, for_cross):
+        d = self.cross_global_args if for_cross else self.global_args
+        return d.get(compiler.get_language(), [])
 
-    def get_project_args(self, compiler, project):
-        args = self.projects_args.get(project)
+    def get_project_args(self, compiler, project, for_cross):
+        d = self.cross_projects_args if for_cross else self.projects_args
+        args = d.get(project)
         if not args:
             return []
         return args.get(compiler.get_language(), [])
 
-    def get_global_link_args(self, compiler):
-        return self.global_link_args.get(compiler.get_language(), [])
+    def get_global_link_args(self, compiler, for_cross):
+        d = self.cross_global_link_args if for_cross else self.global_link_args
+        return d.get(compiler.get_language(), [])
 
-    def get_project_link_args(self, compiler, project):
-        link_args = self.projects_link_args.get(project)
+    def get_project_link_args(self, compiler, project, for_cross):
+        d = self.cross_projects_link_args if for_cross else self.projects_link_args
+
+        link_args = d.get(project)
         if not link_args:
             return []
 
@@ -351,6 +363,8 @@ a hard error in the future.''' % name)
             result[k] = v
         return result
 
+    def is_linkable_target(self):
+        return False
 
 class BuildTarget(Target):
     known_kwargs = known_build_target_kwargs
@@ -657,12 +671,21 @@ class BuildTarget(Target):
     def get_all_link_deps(self):
         return self.get_transitive_link_deps()
 
+    @lru_cache(maxsize=None)
     def get_transitive_link_deps(self):
         result = []
         for i in self.link_targets:
             result += i.get_all_link_deps()
         return result
 
+    @lru_cache(maxsize=None)
+    def get_link_dep_subdirs(self):
+        result = OrderedSet()
+        for i in self.link_targets:
+            result.add(i.get_subdir())
+            result.update(i.get_link_dep_subdirs())
+        return result
+
     def get_custom_install_dir(self):
         return self.install_dir
 
@@ -817,6 +840,14 @@ This will become a hard error in a future Meson release.''')
         self.implicit_include_directories = kwargs.get('implicit_include_directories', True)
         if not isinstance(self.implicit_include_directories, bool):
             raise InvalidArguments('Implicit_include_directories must be a boolean.')
+        self.gnu_symbol_visibility = kwargs.get('gnu_symbol_visibility', '')
+        if not isinstance(self.gnu_symbol_visibility, str):
+            raise InvalidArguments('GNU symbol visibility must be a string.')
+        if self.gnu_symbol_visibility != '':
+            permitted = ['default', 'internal', 'hidden', 'protected', 'inlineshidden']
+            if self.gnu_symbol_visibility not in permitted:
+                raise InvalidArguments('GNU symbol visibility arg %s not one of: %s',
+                                       self.symbol_visibility, ', '.join(permitted))
 
     def get_filename(self):
         return self.filename
@@ -1070,13 +1101,10 @@ You probably should put it in link_with instead.''')
         '''
         linker, _ = self.get_clink_dynamic_linker_and_stdlibs()
         # Mixing many languages with MSVC is not supported yet so ignore stdlibs.
-        if linker and linker.get_id() == 'msvc':
+        if linker and linker.get_id() in ['msvc', 'llvm', 'dmd']:
             return True
         return False
 
-    def is_linkable_target(self):
-        return False
-
     def check_module_linking(self):
         '''
         Warn if shared modules are linked with target: (link_with) #2865
@@ -1295,6 +1323,10 @@ class Executable(BuildTarget):
         # Only linkwithable if using export_dynamic
         self.is_linkwithable = self.export_dynamic
 
+    def description(self):
+        '''Human friendly description of the executable'''
+        return self.name
+
     def type_suffix(self):
         return "@exe"
 
@@ -1373,6 +1405,8 @@ class SharedLibrary(BuildTarget):
     def __init__(self, name, subdir, subproject, is_cross, sources, objects, environment, kwargs):
         self.soversion = None
         self.ltversion = None
+        # Max length 2, first element is compatibility_version, second is current_version
+        self.darwin_versions = []
         self.vs_module_defs = None
         # The import library this target will generate
         self.import_filename = None
@@ -1499,6 +1533,44 @@ class SharedLibrary(BuildTarget):
         self.filename = self.filename_tpl.format(self)
         self.outputs = [self.filename]
 
+    @staticmethod
+    def _validate_darwin_versions(darwin_versions):
+        try:
+            if isinstance(darwin_versions, int):
+                darwin_versions = str(darwin_versions)
+            if isinstance(darwin_versions, str):
+                darwin_versions = 2 * [darwin_versions]
+            if not isinstance(darwin_versions, list):
+                raise InvalidArguments('Shared library darwin_versions: must be a string, integer,'
+                                       'or a list, not {!r}'.format(darwin_versions))
+            if len(darwin_versions) > 2:
+                raise InvalidArguments('Shared library darwin_versions: list must contain 2 or fewer elements')
+            if len(darwin_versions) == 1:
+                darwin_versions = 2 * darwin_versions
+            for i, v in enumerate(darwin_versions[:]):
+                if isinstance(v, int):
+                    v = str(v)
+                if not isinstance(v, str):
+                    raise InvalidArguments('Shared library darwin_versions: list elements '
+                                           'must be strings or integers, not {!r}'.format(v))
+                if not re.fullmatch(r'[0-9]+(\.[0-9]+){0,2}', v):
+                    raise InvalidArguments('Shared library darwin_versions: must be X.Y.Z where '
+                                           'X, Y, Z are numbers, and Y and Z are optional')
+                parts = v.split('.')
+                if len(parts) in (1, 2, 3) and int(parts[0]) > 65535:
+                    raise InvalidArguments('Shared library darwin_versions: must be X.Y.Z '
+                                           'where X is [0, 65535] and Y, Z are optional')
+                if len(parts) in (2, 3) and int(parts[1]) > 255:
+                    raise InvalidArguments('Shared library darwin_versions: must be X.Y.Z '
+                                           'where Y is [0, 255] and Y, Z are optional')
+                if len(parts) == 3 and int(parts[2]) > 255:
+                    raise InvalidArguments('Shared library darwin_versions: must be X.Y.Z '
+                                           'where Z is [0, 255] and Y, Z are optional')
+                darwin_versions[i] = v
+        except ValueError:
+            raise InvalidArguments('Shared library darwin_versions: value is invalid')
+        return darwin_versions
+
     def process_kwargs(self, kwargs, environment):
         super().process_kwargs(kwargs, environment)
 
@@ -1527,6 +1599,13 @@ class SharedLibrary(BuildTarget):
                 # We replicate what Autotools does here and take the first
                 # number of the version by default.
                 self.soversion = self.ltversion.split('.')[0]
+            # macOS and iOS dylib compatibility_version and current_version
+            if 'darwin_versions' in kwargs:
+                self.darwin_versions = self._validate_darwin_versions(kwargs['darwin_versions'])
+            elif self.soversion:
+                # If unspecified, pick the soversion
+                self.darwin_versions = 2 * [self.soversion]
+
         # Visual Studio module-definitions file
         if 'vs_module_defs' in kwargs:
             path = kwargs['vs_module_defs']
@@ -1638,6 +1717,7 @@ class CustomTarget(Target):
         'depfile',
         'build_by_default',
         'override_options',
+        'console',
     ])
 
     def __init__(self, name, subdir, subproject, kwargs, absolute_paths=False):
@@ -1741,7 +1821,7 @@ class CustomTarget(Target):
             if i.strip() == '':
                 raise InvalidArguments('Output must not consist only of whitespace.')
             if has_path_sep(i):
-                raise InvalidArguments('Output must not contain a path segment.')
+                raise InvalidArguments('Output {!r} must not contain a path segment.'.format(i))
             if '@INPUT@' in i or '@INPUT0@' in i:
                 m = 'Output cannot contain @INPUT@ or @INPUT0@, did you ' \
                     'mean @PLAINNAME@ or @BASENAME@?'
@@ -1756,6 +1836,11 @@ class CustomTarget(Target):
         self.capture = kwargs.get('capture', False)
         if self.capture and len(self.outputs) != 1:
             raise InvalidArguments('Capturing can only output to a single file.')
+        self.console = kwargs.get('console', False)
+        if not isinstance(self.console, bool):
+            raise InvalidArguments('"console" kwarg only accepts booleans')
+        if self.capture and self.console:
+            raise InvalidArguments("Can't both capture output and output to console")
         if 'command' not in kwargs:
             raise InvalidArguments('Missing keyword argument "command".')
         if 'depfile' in kwargs:
@@ -1805,7 +1890,8 @@ class CustomTarget(Target):
             while hasattr(ed, 'held_object'):
                 ed = ed.held_object
             if not isinstance(ed, (CustomTarget, BuildTarget)):
-                raise InvalidArguments('Can only depend on toplevel targets: custom_target or build_target (executable or a library)')
+                raise InvalidArguments('Can only depend on toplevel targets: custom_target or build_target (executable or a library) got: %s(%s)'
+                                       % (type(ed), ed))
             self.extra_depends.append(ed)
         for i in depend_files:
             if isinstance(i, (File, str)):
@@ -1911,6 +1997,9 @@ class Jar(BuildTarget):
         for s in self.sources:
             if not s.endswith('.java'):
                 raise InvalidArguments('Jar source %s is not a java file.' % s)
+        for t in self.link_targets:
+            if not isinstance(t, Jar):
+                raise InvalidArguments('Link target %s is not a jar target.' % t)
         self.filename = self.name + '.jar'
         self.outputs = [self.filename]
         self.java_args = kwargs.get('java_args', [])
@@ -1928,6 +2017,13 @@ class Jar(BuildTarget):
         # All jar targets are installable.
         pass
 
+    def is_linkable_target(self):
+        return True
+
+    def get_classpath_args(self):
+        cp_paths = [os.path.join(l.get_subdir(), l.get_filename()) for l in self.link_targets]
+        return ['-cp', os.pathsep.join(cp_paths)]
+
 class CustomTargetIndex:
 
     """A special opaque object returned by indexing a CustomTarget. This object
@@ -1950,7 +2046,6 @@ class CustomTargetIndex:
     def get_subdir(self):
         return self.target.get_subdir()
 
-
 class ConfigureFile:
 
     def __init__(self, subdir, sourcename, targetname, configuration_data):
